# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import concurrent.futures
import os
import shutil
import subprocess
import tempfile
from typing import Dict, List, Optional

import click
import platformdirs
import requests
import yaml
from diskcache import FanoutCache

from magnum_cluster_api import image_utils

CACHE = FanoutCache(
    directory=platformdirs.user_cache_dir("magnum_cluster_api", "vexxhost"),
)

# NOTE(mnaser): This is a list of all the Kubernetes versions which we've
#               released images for.  This list is used to determine which
#               images of Kubernetes we should publish to the registry.
VERSIONS = [
    "v1.23.13",
    "v1.23.17",
    "v1.24.7",
    "v1.24.15",
    "v1.25.3",
    "v1.25.11",
    "v1.26.2",
    "v1.26.6",
    "v1.26.11",
    "v1.27.3",
    "v1.27.8",
    "v1.27.15",
    "v1.28.11",
    "v1.29.6",
    "v1.30.2",
    "v1.31.1",
]


@click.command()
@click.option(
    "--repository",
    required=True,
    help="Target image repository",
)
@click.option(
    "--config",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=str),
    help="Path to YAML config file containing image configurations",
)
@click.option(
    "--parallel",
    default=8,
    help="Number of parallel uploads",
)
@click.option(
    "--insecure",
    is_flag=True,
    help="Allow insecure connections to the registry.",
)
@click.option(
    "--platforms",
    default="linux/amd64",
    help="Comma-separated list of platforms to copy (e.g., linux/amd64,linux/arm64)",
)
def main(repository, config, parallel, insecure, platforms):
    """
    Load images into a remote registry for `container_infra_prefix` usage.
    """

    platforms = platforms.split(",")

    config_data = load_config(config) if config else None

    # NOTE(mnaser): This list must be maintained manually because the image
    #               registry must be able to support a few different versions
    #               of Kubernetes since it is possible to have multiple
    #               clusters running different versions of Kubernetes at the
    #               same time.
    images = set(
        _get_all_kubeadm_images(config_data)
        + _get_calico_images(config_data)
        + _get_cilium_images(config_data)
        + _get_cloud_provider_images(config_data)
        + _get_infra_images(config_data)
    )

    crane_path = shutil.which("crane")

    if crane_path is None:
        raise click.UsageError(
            """Crane is not installed. Please install it before running this command:
             https://github.com/google/go-containerregistry/blob/main/cmd/crane/README.md#installation"""
        )

    with concurrent.futures.ThreadPoolExecutor(max_workers=parallel) as executor:
        future_to_image = {
            executor.submit(
                _mirror_image, image, repository, insecure, crane_path, platforms
            ): image
            for image in images
        }

        for future in concurrent.futures.as_completed(future_to_image):
            image = future_to_image[future]
            try:
                future.result()
            except Exception as e:
                click.echo(
                    f"Image upload failed for {image}: {e}",
                    err=True,
                )


def _mirror_image(
    image: str, repository: str, insecure: bool, crane_path: str, platforms: List[str]
):
    """Mirror image to target registry with multi-arch support."""
    src = image
    dst = image_utils.get_image(image, repository)
    version = src.split(":")[-1]
    temp_images = []

    try:
        # Check which platforms are supported
        supported_platforms = []
        for platform in platforms:
            if _check_platform_support(src, crane_path, platform, insecure):
                supported_platforms.append(platform)
                click.echo(f"Platform {platform} is supported for {src}")
            else:
                click.echo(f"Platform {platform} is not supported for {src}", err=True)

        if not supported_platforms:
            click.echo(f"No supported platforms found for {src}", err=True)
            return

        # Copy each supported platform variant
        for platform in supported_platforms:
            platform_tag = f"{version}-{platform.replace('/', '-')}"
            temp_dst = dst.replace(f":{version}", f":{platform_tag}")

            command = [crane_path]
            if insecure:
                command.append("--insecure")
            command += ["copy", "--platform", platform, src, temp_dst]

            try:
                subprocess.run(
                    command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                )
                click.echo(f"Copied {platform} variant of {src}")
                temp_images.append(temp_dst)
            except subprocess.CalledProcessError as e:
                click.echo(f"Failed to copy {platform} variant of {src}: {e}", err=True)
                continue

        # Create multi-arch manifest if we have platform variants
        if len(temp_images) > 0:
            try:
                command = [crane_path]
                if insecure:
                    command.append("--insecure")
                command += ["index", "append"]

                # Add each platform manifest
                for temp_image in temp_images:
                    command.extend(["--manifest", temp_image])

                # Add the target tag
                command.extend(["--tag", dst])

                subprocess.run(
                    command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                )
                click.echo(f"Created multi-arch manifest for {src}")

            except subprocess.CalledProcessError as e:
                click.echo(f"Failed to create manifest list: {e}", err=True)

    except Exception as e:
        click.echo(f"Error processing {src}: {e}", err=True)
        return


def _get_all_kubeadm_images(config: Optional[Dict] = None):
    """
    Get the list of images that are used by Kubernetes by downloading "kubeadm"
    and running the "kubeadm config images list" command.
    """

    images = []
    versions = VERSIONS
    if config and "kubernetes" in config:
        versions = config["kubernetes"]["versions"]

    for version in versions:
        images += _get_kubeadm_images(version)

    return images


@CACHE.memoize()
def _get_kubeadm_images(version: str):
    """
    Get the list of images that are used by Kubernetes by downloading "kubeadm"
    and running the "kubeadm config images list" command.
    """

    # Download kubeadm
    r = requests.get(f"https://dl.k8s.io/release/{version}/bin/linux/amd64/kubeadm")
    r.raise_for_status()

    # Write it to a temporary file
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(r.content)
        f.close()

        # Make it executable
        os.chmod(f.name, 0o755)

        # Run the command
        output = subprocess.check_output(
            [f.name, "config", "images", "list", "--kubernetes-version", version]
        )

        # Remove the temporary file
        os.unlink(f.name)

    # Parse the output
    return output.decode().replace("k8s.gcr.io", "registry.k8s.io").splitlines()


def _get_calico_images(config: Optional[Dict] = None) -> List[str]:
    if config and "calico" in config:
        images = []
        for image in config["calico"]["images"]:
            images.extend([f"{image['name']}:{tag}" for tag in image["tags"]])
        return images

    return [
        # v3.24.2
        "quay.io/calico/cni:v3.24.2",
        "quay.io/calico/kube-controllers:v3.24.2",
        "quay.io/calico/node:v3.24.2",
        # v3.25.2
        "quay.io/calico/cni:v3.25.2",
        "quay.io/calico/kube-controllers:v3.25.2",
        "quay.io/calico/node:v3.25.2",
        # v3.26.5
        "quay.io/calico/cni:v3.26.5",
        "quay.io/calico/kube-controllers:v3.26.5",
        "quay.io/calico/node:v3.26.5",
        # v3.27.4
        "quay.io/calico/cni:v3.27.4",
        "quay.io/calico/kube-controllers:v3.27.4",
        "quay.io/calico/node:v3.27.4",
        # v3.28.2
        "quay.io/calico/cni:v3.28.2",
        "quay.io/calico/kube-controllers:v3.28.2",
        "quay.io/calico/node:v3.28.2",
        # v3.29.0
        "quay.io/calico/cni:v3.29.0",
        "quay.io/calico/kube-controllers:v3.29.0",
        "quay.io/calico/node:v3.29.0",
        # v3.29.2
        "quay.io/calico/cni:v3.29.2",
        "quay.io/calico/kube-controllers:v3.29.2",
        "quay.io/calico/node:v3.29.2",
        # v3.29.3
        "quay.io/calico/cni:v3.29.3",
        "quay.io/calico/kube-controllers:v3.29.3",
        "quay.io/calico/node:v3.29.3",
        # v3.30.0
        "quay.io/calico/cni:v3.30.0",
        "quay.io/calico/kube-controllers:v3.30.0",
        "quay.io/calico/node:v3.30.0",
    ]


def _get_cilium_images(config: Optional[Dict] = None) -> List[str]:
    if config and "cilium" in config:
        images = []
        for image in config["cilium"]["images"]:
            images.extend([f"{image['name']}:{tag}" for tag in image["tags"]])
        return images

    return [
        # v1.15.3
        "quay.io/cilium/cilium:v1.15.3",
        "quay.io/cilium/operator-generic:v1.15.3",
        # v1.15.6
        "quay.io/cilium/cilium:v1.15.6",
        "quay.io/cilium/operator-generic:v1.15.6",
    ]


def _get_cloud_provider_images(config: Optional[Dict] = None) -> List[str]:
    if config and "cloud_provider" in config:
        images = []
        for image in config["cloud_provider"]["images"]:
            images.extend([f"{image['name']}:{tag}" for tag in image["tags"]])
        return images

    return [
        # v1.24.6
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.24.6",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.24.6",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.24.6",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.24.6",
        # v1.25.3
        "docker.io/k8scloudprovider/k8s-keystone-auth:v1.25.3",
        "docker.io/k8scloudprovider/cinder-csi-plugin:v1.25.3",
        "docker.io/k8scloudprovider/manila-csi-plugin:v1.25.3",
        "docker.io/k8scloudprovider/openstack-cloud-controller-manager:v1.25.3",
        # v1.25.6
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.25.6",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.25.6",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.25.6",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.25.6",
        # v1.26.3
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.26.3",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.26.3",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.26.3",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.26.3",
        # v1.27.2
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.27.2",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.27.2",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.27.2",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.27.2",
        # v1.27.3
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.27.3",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.27.3",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.27.3",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.27.3",
        # v1.28.0
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.28.0",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.28.0",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.28.0",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.28.0",
        # v1.28.2
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.28.2",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.28.2",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.28.2",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.28.2",
        # v1.29.0
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.29.0",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.29.0",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.29.0",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.29.0",
        # v1.30.0
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.30.0",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.30.0",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.30.0",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.30.0",
        # v1.31.1
        "registry.k8s.io/provider-os/k8s-keystone-auth:v1.31.1",
        "registry.k8s.io/provider-os/cinder-csi-plugin:v1.31.1",
        "registry.k8s.io/provider-os/manila-csi-plugin:v1.31.1",
        "registry.k8s.io/provider-os/openstack-cloud-controller-manager:v1.31.1",
    ]


def _get_infra_images(config: Optional[Dict] = None) -> List[str]:
    if config and "infra" in config:
        images = []
        for image in config["infra"]["images"]:
            images.extend([f"{image['name']}:{tag}" for tag in image["tags"]])
        return images

    return [
        "registry.k8s.io/sig-storage/csi-attacher:v3.4.0",
        "registry.k8s.io/sig-storage/csi-attacher:v4.2.0",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.4.0",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.5.1",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.6.2",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.6.3",
        "registry.k8s.io/sig-storage/csi-node-driver-registrar:v2.9.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.0.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.1.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.3.0",
        "registry.k8s.io/sig-storage/csi-provisioner:v3.4.1",
        "registry.k8s.io/sig-storage/csi-resizer:v1.4.0",
        "registry.k8s.io/sig-storage/csi-resizer:v1.8.0",
        "registry.k8s.io/sig-storage/csi-snapshotter:v5.0.1",
        "registry.k8s.io/sig-storage/csi-snapshotter:v6.0.1",
        "registry.k8s.io/sig-storage/csi-snapshotter:v6.2.1",
        "registry.k8s.io/sig-storage/livenessprobe:v2.7.0",
        "registry.k8s.io/sig-storage/livenessprobe:v2.8.0",
        "registry.k8s.io/sig-storage/livenessprobe:v2.9.0",
        "registry.k8s.io/sig-storage/nfsplugin:v4.2.0",
    ]


def load_config(config_path: str) -> Optional[Dict]:
    """
    Load configuration from YAML file. Missing keys will use default values.
    """

    try:
        with open(config_path, "r") as f:
            config = yaml.safe_load(f)

        if not isinstance(config, dict):
            raise click.UsageError("Config file must contain a YAML dictionary")

        if "kubernetes" in config:
            if not isinstance(config["kubernetes"].get("versions", []), list):
                raise click.UsageError("kubernetes.versions must be a list")

        for component in ["calico", "cilium", "cloud_provider", "infra"]:
            if component in config:
                if not isinstance(config[component].get("images", []), list):
                    raise click.UsageError(f"{component}.images must be a list")

                for image in config[component].get("images", []):
                    if not isinstance(image.get("name", ""), str):
                        raise click.UsageError(
                            f"{component} image name must be a string"
                        )
                    if not isinstance(image.get("tags", []), list):
                        raise click.UsageError(f"{component} image tags must be a list")

        return config

    except yaml.YAMLError as e:
        raise click.UsageError(f"Failed to parse config file: {e}")
    except OSError as e:
        raise click.UsageError(f"Failed to read config file: {e}")


def _check_platform_support(
    image: str, crane_path: str, platform: str, insecure: bool
) -> bool:
    """Check if image supports specific platform."""
    try:
        command = [crane_path]
        if insecure:
            command.append("--insecure")
        command += ["manifest", "--platform", platform, image]
        subprocess.check_output(command, stderr=subprocess.DEVNULL)
        return True
    except subprocess.CalledProcessError:
        return False



from diskcache import Cache  # type: ignore


class ServerGroupCache(object):
    """
    Cache server group ids by project id and server group name

    It uses diskcache to cache the server group ids. We can replace it with dogpile or whatever
    if we need.
    """

    def __init__(self, ttl: int = 300):
        self._cache = Cache()
        self._ttl = ttl

    def get(self, project_id, server_group_name):
        return self._cache.get((project_id, server_group_name))

    def set(self, project_id, server_group_name, server_group_id):
        self._cache.set((project_id, server_group_name), server_group_id, self._ttl)



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import semver

from magnum_cluster_api import conf

CONF = conf.CONF

PAUSE = "registry.k8s.io/pause:3.9"

CLUSTER_AUTOSCALER_LATEST_BY_MINOR = {
    "1.22": "1.22.3",
    "1.23": "1.23.1",
    "1.24": "1.24.3",
    "1.25": "1.25.3",
    "1.26": "1.26.8",
    "1.27": "1.27.8",
    "1.28": "1.28.7",
    "1.29": "1.29.5",
    "1.30": "1.30.5",
    "1.31": "1.31.3",
    "1.32": "1.32.2",
    "1.33": "1.33.0",
}


def get_cluster_autoscaler_image(version: str):
    parsed_version = semver.VersionInfo.parse(version[1:])
    cluster_autoscaler_version = CLUSTER_AUTOSCALER_LATEST_BY_MINOR.get(
        f"{parsed_version.major}.{parsed_version.minor}",
        f"{parsed_version.major}.{parsed_version.minor}.0",
    )

    return f"{CONF.auto_scaling.image_repository}/cluster-autoscaler:v{cluster_autoscaler_version}"



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import glob
import math
import os
import types
import typing

import certifi
import pkg_resources
import pykube  # type: ignore
import yaml
from magnum import objects as magnum_objects  # type: ignore
from magnum.common import context, neutron  # type: ignore
from magnum.common import utils as magnum_utils  # type: ignore
from magnum.common.cert_manager import cert_manager  # type: ignore
from magnum.common.x509 import operations as x509  # type: ignore
from magnum.conductor.handlers.common import (
    cert_manager as cert_manager_handlers,  # type: ignore
)
from oslo_config import cfg  # type: ignore
from oslo_serialization import base64  # type: ignore
from oslo_utils import encodeutils  # type: ignore

from magnum_cluster_api import (
    clients,
    helm,
    image_utils,
    images,
    magnum_cluster_api,
    objects,
    utils,
)
from magnum_cluster_api.integrations import cinder, manila

CONF = cfg.CONF
CALICO_TAG = "v3.30.2"

CLUSTER_CLASS_VERSION = pkg_resources.require("magnum_cluster_api")[0].version
CLUSTER_CLASS_NAME = f"magnum-v{CLUSTER_CLASS_VERSION}"
CLUSTER_CLASS_NODE_VOLUME_DETACH_TIMEOUT = "300s"  # seconds

AUTOSCALE_ANNOTATION_MIN = "cluster.x-k8s.io/cluster-api-autoscaler-node-group-min-size"
AUTOSCALE_ANNOTATION_MAX = "cluster.x-k8s.io/cluster-api-autoscaler-node-group-max-size"

DEFAULT_POD_CIDR = "10.100.0.0/16"


class ClusterAutoscalerHelmRelease:
    def __init__(self, api, cluster) -> None:
        self.cluster = cluster

    @property
    def apply(self):
        image = images.get_cluster_autoscaler_image(
            utils.get_kube_tag(self.cluster),
        )
        image_repo, image_tag = image.split(":", 1)

        return helm.UpgradeReleaseCommand(
            namespace="magnum-system",
            release_name=self.cluster.stack_id,
            chart_ref=os.path.join(
                pkg_resources.resource_filename("magnum_cluster_api", "charts"),
                "cluster-autoscaler/",
            ),
            values={
                "fullnameOverride": f"{self.cluster.stack_id}-autoscaler",
                "cloudProvider": "clusterapi",
                "clusterAPIMode": "kubeconfig-incluster",
                "clusterAPIKubeconfigSecret": f"{self.cluster.stack_id}-kubeconfig",
                "autoDiscovery": {
                    "clusterName": self.cluster.stack_id,
                },
                "image": {
                    "repository": image_repo,
                    "tag": image_tag,
                },
                "nodeSelector": {
                    "openstack-control-plane": "enabled",
                },
                "extraArgs": {
                    "logtostderr": True,
                    "stderrthreshold": "info",
                    "v": 4,
                    "enforce-node-group-min-size": True,
                },
            },
        )

    @property
    def delete(self):
        return helm.DeleteReleaseCommand(
            namespace="magnum-system",
            release_name=self.cluster.stack_id,
            skip_missing=True,
        )


class ClusterServerGroups:
    def __init__(
        self, context: context.RequestContext, cluster: magnum_objects.Cluster
    ) -> None:
        self.cluster = cluster
        self.context = context
        self.osc = clients.get_openstack_api(self.context)

    def apply(self):
        # Create a server group for controlplane
        utils.ensure_controlplane_server_group(ctx=self.context, cluster=self.cluster)

        # Create a server group per a nodegroup
        for ng in self.cluster.nodegroups:
            if ng.role == "master":
                continue
            utils.ensure_worker_server_group(
                ctx=self.context, cluster=self.cluster, node_group=ng
            )

    def delete(self):
        # delete controlplane server group
        utils.delete_controlplane_server_group(ctx=self.context, cluster=self.cluster)

        # Create worker server groups
        for ng in self.cluster.nodegroups:
            if ng.role == "master":
                continue

            utils.delete_worker_server_group(
                ctx=self.context, cluster=self.cluster, node_group=ng
            )


class Base(abc.ABC):
    def __init__(self, api: magnum_cluster_api.KubeClient, namespace="magnum-system"):
        self.api = api
        self.namespace = namespace

    @property
    @abc.abstractmethod
    def api_version(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def kind(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @abc.abstractmethod
    def get_object(self) -> dict:
        pass

    def get_resource(self) -> dict:
        resource = self.get_object()

        resource["apiVersion"] = self.api_version
        resource["kind"] = self.kind

        resource.setdefault("metadata", {})
        resource["metadata"].setdefault("name", self.name)
        if self.kind != "Namespace":
            resource["metadata"].setdefault("namespace", self.namespace)

        return resource

    def apply(self):
        resource = self.get_resource()
        self.api.create_or_update(resource)

    def delete(self) -> None:
        self.api.delete(
            self.api_version, self.kind, self.name, namespace=self.namespace
        )


class ClusterBase(Base):
    def __init__(
        self, api: magnum_cluster_api.KubeClient, cluster: magnum_objects.Cluster
    ):
        super().__init__(api)
        self.cluster = cluster

    @property
    def name(self) -> str:
        return self.cluster.uuid

    @property
    def labels(self) -> dict:
        return {
            "cluster-uuid": self.cluster.uuid,
        }


class CloudProviderClusterResourcesSecret(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: magnum_cluster_api.KubeClient,
        pykube_api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
        namespace: str = "magnum-system",
    ):
        self.context = context
        self.api = api
        self.pykube_api = pykube_api
        self.cluster = cluster
        self.namespace = namespace

    @property
    def api_version(self) -> str:
        return "v1"

    @property
    def kind(self) -> str:
        return "Secret"

    @property
    def name(self) -> str:
        return f"{self.cluster.stack_id}-cloud-provider"

    def get_object(self) -> dict:
        data = {
            **magnum_cluster_api.Driver.get_cloud_provider_cluster_resource_secret_data(
                self.cluster
            ),
            **{
                "cloud-config-secret.yaml": yaml.dump(
                    {
                        "apiVersion": pykube.Secret.version,
                        "kind": pykube.Secret.kind,
                        "metadata": {
                            "name": "cloud-config",
                            "namespace": "kube-system",
                        },
                        "stringData": {
                            "cloud.conf": utils.generate_cloud_controller_manager_config(
                                self.context,
                                self.pykube_api,
                                self.cluster,
                            ),
                            "ca.crt": magnum_utils.get_openstack_ca(),
                        },
                    }
                ),
            },
        }

        osc = clients.get_openstack_api(self.context)
        if cinder.is_enabled(self.cluster):
            volume_types = osc.cinder().volume_types.list()
            default_volume_type = osc.cinder().volume_types.default()
            data = {
                **data,
                **magnum_cluster_api.Driver.get_cinder_csi_cluster_resource_secret_data(
                    self.cluster
                ),
                **{
                    f"storageclass-block-{vt.name}.yaml": yaml.dump(
                        {
                            "apiVersion": objects.StorageClass.version,
                            "allowVolumeExpansion": True,
                            "kind": objects.StorageClass.kind,
                            "metadata": {
                                "annotations": (
                                    {
                                        "storageclass.kubernetes.io/is-default-class": "true"
                                    }
                                    if default_volume_type.name == vt.name
                                    else {}
                                ),
                                "name": "block-%s" % utils.convert_to_rfc1123(vt.name),
                            },
                            "provisioner": "cinder.csi.openstack.org",
                            "parameters": {
                                "type": vt.name,
                            },
                            "reclaimPolicy": "Delete",
                            "volumeBindingMode": "Immediate",
                        }
                    )
                    for vt in volume_types
                    if vt.name != "__DEFAULT__"
                },
            }

        if manila.is_enabled(self.cluster):
            share_types = osc.manila().share_types.list()
            share_network_id = self.cluster.labels.get("manila_csi_share_network_id")
            data = {
                **data,
                **magnum_cluster_api.Driver.get_manila_csi_cluster_resource_secret_data(
                    self.cluster
                ),
                **{
                    "manila-csi-secret.yaml": yaml.dump(
                        {
                            "apiVersion": pykube.Secret.version,
                            "kind": pykube.Secret.kind,
                            "metadata": {
                                "name": "csi-manila-secrets",
                                "namespace": "kube-system",
                            },
                            "stringData": utils.generate_manila_csi_cloud_config(
                                self.context,
                                self.pykube_api,
                                self.cluster,
                            ),
                        },
                    )
                },
            }
            # NOTE: We only create StorageClasses if share_network_id specified.
            if share_network_id:
                data = {
                    **data,
                    **{
                        f"storageclass-share-{st.name}.yaml": yaml.dump(
                            {
                                "apiVersion": objects.StorageClass.version,
                                "allowVolumeExpansion": True,
                                "kind": objects.StorageClass.kind,
                                "metadata": {
                                    "name": "share-%s"
                                    % utils.convert_to_rfc1123(st.name),
                                },
                                "provisioner": "nfs.manila.csi.openstack.org",
                                "parameters": {
                                    "type": st.name,
                                    "shareNetworkID": share_network_id,
                                    "csi.storage.k8s.io/provisioner-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/provisioner-secret-namespace": "kube-system",
                                    "csi.storage.k8s.io/controller-expand-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/controller-expand-secret-namespace": "kube-system",
                                    "csi.storage.k8s.io/node-stage-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/node-stage-secret-namespace": "kube-system",
                                    "csi.storage.k8s.io/node-publish-secret-name": "csi-manila-secrets",
                                    "csi.storage.k8s.io/node-publish-secret-namespace": "kube-system",
                                },
                                "reclaimPolicy": "Delete",
                                "volumeBindingMode": "Immediate",
                            }
                        )
                        for st in share_types
                    },
                }

        return {
            "type": "addons.cluster.x-k8s.io/resource-set",
            "stringData": data,
        }

    def get_or_none(self) -> objects.Cluster:
        return pykube.Secret.objects(
            self.pykube_api, namespace=self.namespace
        ).get_or_none(name=self.name)

    def delete(self):
        cr_cm = self.get_or_none()
        if cr_cm:
            cr_cm.delete()


class LegacyClusterResourcesSecret(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: magnum_cluster_api.KubeClient,
        pykube_api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
        namespace: str = "magnum-system",
    ):
        self.context = context
        self.api = api
        self.pykube_api = pykube_api
        self.cluster = cluster
        self.namespace = namespace

    @property
    def api_version(self) -> str:
        return "v1"

    @property
    def kind(self) -> str:
        return "Secret"

    def get_object(self) -> dict:
        repository = utils.get_cluster_container_infra_prefix(self.cluster)
        manifests_path = pkg_resources.resource_filename(
            "magnum_cluster_api", "manifests"
        )

        data = magnum_cluster_api.Driver.get_legacy_cluster_resource_secret_data(
            self.cluster
        )

        data = {
            **data,
            **{
                "cloud-config-secret.yaml": yaml.dump(
                    {
                        "apiVersion": pykube.Secret.version,
                        "kind": pykube.Secret.kind,
                        "metadata": {
                            "name": "cloud-config",
                            "namespace": "kube-system",
                        },
                        "stringData": {
                            "cloud.conf": utils.generate_cloud_controller_manager_config(
                                self.context,
                                self.pykube_api,
                                self.cluster,
                            ),
                            "ca.crt": magnum_utils.get_openstack_ca(),
                        },
                    }
                ),
            },
        }

        if self.cluster.cluster_template.network_driver == "calico":
            calico_version = self.cluster.labels.get("calico_tag", CALICO_TAG)
            data = {
                **data,
                **{
                    "calico.yml": image_utils.update_manifest_images(
                        self.cluster.uuid,
                        os.path.join(manifests_path, f"calico/{calico_version}.yaml"),
                        repository=repository,
                    )
                },
            }

        if manila.is_enabled(self.cluster):
            data = {
                **data,
                **{
                    os.path.basename(manifest): image_utils.update_manifest_images(
                        self.cluster.uuid,
                        manifest,
                        repository=repository,
                    )
                    for manifest in glob.glob(
                        os.path.join(manifests_path, "nfs-csi/*.yaml")
                    )
                },
            }

        osc = clients.get_openstack_api(self.context)
        if utils.get_cluster_label_as_bool(self.cluster, "keystone_auth_enabled", True):
            auth_url = osc.url_for(
                service_type="identity",
                interface="public",
            )
            data = {
                **data,
                **{
                    "keystone-auth.yaml": helm.TemplateReleaseCommand(
                        namespace="kube-system",
                        release_name="k8s-keystone-auth",
                        chart_ref=os.path.join(
                            pkg_resources.resource_filename(
                                "magnum_cluster_api", "charts"
                            ),
                            "k8s-keystone-auth/",
                        ),
                        values={
                            "conf": {
                                "auth_url": auth_url
                                + ("" if auth_url.endswith("/v3") else "/v3"),
                                "ca_cert": magnum_utils.get_openstack_ca(),
                                "policy": utils.get_keystone_auth_default_policy(
                                    self.cluster
                                ),
                            },
                        },
                    )(repository=repository)
                },
            }

        return {
            "type": "addons.cluster.x-k8s.io/resource-set",
            "stringData": data,
        }

    def get_or_none(self) -> objects.Cluster:
        return pykube.Secret.objects(
            self.pykube_api, namespace="magnum-system"
        ).get_or_none(name=self.name)

    def delete(self):
        cr_cm = self.get_or_none()
        if cr_cm:
            cr_cm.delete()


class CertificateAuthoritySecret(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: magnum_cluster_api.KubeClient,
        pykube_api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
    ):
        super().__init__(api, cluster)
        self.pykube_api = pykube_api
        self.context = context

    @abc.abstractmethod
    def certificate_name(self) -> str:
        pass

    @abc.abstractmethod
    def magnum_cert_ref_name(self) -> str:
        pass

    @property
    def api_version(self) -> str:
        return "v1"

    @property
    def kind(self) -> str:
        return "Secret"

    @property
    def name(self) -> str:
        return f"{self.cluster.stack_id}-{self.certificate_name()}"

    def delete(self) -> None:
        resource = self.get_or_none()
        if resource:
            resource.delete()

    def get_or_none(self) -> pykube.Secret:
        return pykube.Secret.objects(
            self.pykube_api, namespace="magnum-system"
        ).get_or_none(name=self.name)

    def get_certificate(self) -> cert_manager.Cert:
        raise NotImplementedError()

    def get_object(self) -> dict:
        cert_ref = getattr(self.cluster, self.magnum_cert_ref_name())
        if cert_ref is None:
            raise Exception(
                "Certificate for %s doesn't exist." % self.magnum_cert_ref_name()
            )
        ca_cert = self.get_certificate()

        return {
            "type": "cluster.x-k8s.io/secret",
            "metadata": {
                "labels": {
                    "cluster.x-k8s.io/cluster-name": f"{self.cluster.stack_id}",
                },
            },
            "stringData": {
                "tls.crt": encodeutils.safe_decode(ca_cert.get_certificate()),
                "tls.key": encodeutils.safe_decode(
                    x509.decrypt_key(
                        ca_cert.get_private_key(),
                        ca_cert.get_private_key_passphrase(),
                    )
                ),
            },
        }


class ApiCertificateAuthoritySecret(CertificateAuthoritySecret):
    def certificate_name(self):
        return "ca"

    def magnum_cert_ref_name(self):
        return "ca_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_ca_certificate(
            self.cluster, self.context, "kubernetes"
        )


class EtcdCertificateAuthoritySecret(CertificateAuthoritySecret):
    def certificate_name(self):
        return "etcd"

    def magnum_cert_ref_name(self):
        return "etcd_ca_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_ca_certificate(
            self.cluster, self.context, "etcd"
        )


class FrontProxyCertificateAuthoritySecret(CertificateAuthoritySecret):
    def certificate_name(self):
        return "proxy"

    def magnum_cert_ref_name(self):
        return "front_proxy_ca_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_ca_certificate(
            self.cluster, self.context, "front-proxy"
        )


class ServiceAccountCertificateAuthoritySecret(CertificateAuthoritySecret):
    def certificate_name(self):
        return "sa"

    def magnum_cert_ref_name(self):
        return "magnum_cert_ref"

    def get_certificate(self) -> cert_manager.Cert:
        return cert_manager_handlers.get_cluster_magnum_cert(self.cluster, self.context)


class CloudConfigSecret(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: magnum_cluster_api.KubeClient,
        cluster: magnum_objects.Cluster,
        region_name: typing.Optional[str] = None,
        credential: types.SimpleNamespace = types.SimpleNamespace(id=None, secret=None),
    ):
        super().__init__(api, cluster)
        self.context = context
        osc = clients.get_openstack_api(self.context)
        self.auth_url = osc.url_for(
            service_type="identity",
            interface=CONF.capi_client.endpoint_type.replace("URL", ""),
        )
        self.region_name = region_name
        self.credential = credential

    @property
    def api_version(self) -> str:
        return "v1"

    @property
    def kind(self) -> str:
        return "Secret"

    @property
    def name(self) -> str:
        return utils.get_cluster_api_cloud_config_secret_name(self.cluster)

    def get_object(self) -> dict:
        ca_certificate = utils.get_capi_client_ca_cert()

        return {
            "metadata": {
                "labels": self.labels,
            },
            "stringData": {
                "cacert": (
                    ca_certificate
                    if ca_certificate
                    else open(certifi.where(), "r").read()
                ),
                "clouds.yaml": yaml.dump(
                    {
                        "clouds": {
                            "default": {
                                "region_name": self.region_name,
                                "endpoint_type": CONF.capi_client.endpoint_type.replace(
                                    "URL", ""
                                ),
                                "identity_api_version": 3,
                                "verify": not CONF.capi_client.insecure,
                                "auth": {
                                    "auth_url": self.auth_url,
                                    "application_credential_id": self.credential.id,
                                    "application_credential_secret": self.credential.secret,
                                },
                            }
                        }
                    }
                ),
            },
        }


def mutate_machine_deployment(
    context: context.RequestContext,
    cluster: objects.Cluster,
    node_group: magnum_objects.NodeGroup,
    machine_deployment: dict = None,
):
    """
    This function will either makes updates to machine deployment fields which
    will not cause a rolling update or will return a new machine deployment
    if none is provided.
    """

    # NOTE(okozachenko1203): Initialize as an empty dict if not provided
    #                        instead of using mutable default argument.
    if machine_deployment is None:
        machine_deployment = {}

    auto_scaling_enabled = utils.get_auto_scaling_enabled(cluster)

    machine_deployment.setdefault(
        "metadata",
        {
            "annotations": {},
            "labels": {},
        },
    )

    # Node labels
    machine_deployment["metadata"]["labels"] = {
        f"node-role.kubernetes.io/{node_group.role}": "",
        "node.cluster.x-k8s.io/nodegroup": node_group.name,
    }

    # Lookup the node group resources
    osc = clients.get_openstack_api(context)
    flavor = utils.lookup_flavor(osc, node_group.flavor_id)
    image = utils.lookup_image(osc, node_group.image_id)

    # Replicas (or min/max if auto-scaling is enabled)
    if auto_scaling_enabled:
        boot_volume_size = utils.get_node_group_label_as_int(
            node_group,
            "boot_volume_size",
            CONF.cinder.default_boot_volume_size,
        )
        if boot_volume_size == 0:
            boot_volume_size = flavor.disk

        machine_deployment["replicas"] = None
        machine_deployment["metadata"]["annotations"] = {
            AUTOSCALE_ANNOTATION_MIN: str(node_group.min_node_count),
            AUTOSCALE_ANNOTATION_MAX: str(
                utils.get_node_group_max_node_count(node_group)
            ),
            "capacity.cluster-autoscaler.kubernetes.io/memory": f"{math.ceil(flavor.ram / 1024)}G",
            "capacity.cluster-autoscaler.kubernetes.io/cpu": str(flavor.vcpus),
            "capacity.cluster-autoscaler.kubernetes.io/ephemeral-disk": str(
                boot_volume_size
            ),
        }
    else:
        machine_deployment["replicas"] = node_group.node_count
        machine_deployment["metadata"]["annotations"] = {}

    # Fixes
    machine_deployment["nodeVolumeDetachTimeout"] = (
        CLUSTER_CLASS_NODE_VOLUME_DETACH_TIMEOUT
    )

    # Anything beyond this point will *NOT* be changed in the machine deployment
    # for update operations (i.e. if the machine deployment already exists).
    if machine_deployment.get("name") == node_group.name:
        return machine_deployment

    # At this point, this is all code that will be added for brand-new machine
    # deployments.  We can bring any of this code into the above block if we
    # want to change it for existing machine deployments.

    machine_deployment.update(
        {
            "class": "default-worker",
            "name": node_group.name,
            "failureDomain": node_group.labels.get("availability_zone"),
            "machineHealthCheck": {"enable": utils.get_auto_healing_enabled(cluster)},
            "variables": {
                "overrides": [
                    {
                        "name": "bootVolume",
                        "value": {
                            "size": utils.get_node_group_label_as_int(
                                node_group,
                                "boot_volume_size",
                                CONF.cinder.default_boot_volume_size,
                            ),
                            "type": node_group.labels.get(
                                "boot_volume_type",
                                cinder.get_default_boot_volume_type(context),
                            ),
                        },
                    },
                    {
                        "name": "flavor",
                        "value": flavor.name,
                    },
                    {
                        "name": "imageRepository",
                        "value": node_group.labels.get(
                            "container_infra_prefix",
                            "",
                        ),
                    },
                    {
                        "name": "imageUUID",
                        "value": image.get("id"),
                    },
                    {
                        "name": "hardwareDiskBus",
                        "value": image.get("hw_disk_bus", ""),
                    },
                    # NOTE(oleks): Override using MachineDeployment-level variables for node groups
                    {
                        "name": "serverGroupId",
                        "value": utils.ensure_worker_server_group(
                            ctx=context, cluster=cluster, node_group=node_group
                        ),
                    },
                    {
                        "name": "isServerGroupDiffFailureDomain",
                        "value": utils.is_node_group_different_failure_domain(
                            node_group=node_group, cluster=cluster
                        ),
                    },
                ],
            },
        }
    )
    return machine_deployment


def generate_machine_deployments_for_cluster(
    context: context.RequestContext, cluster: objects.Cluster
) -> list:
    machine_deployments = []
    for ng in cluster.nodegroups:
        if ng.role == "master" or ng.status.startswith("DELETE"):
            continue

        machine_deployment = mutate_machine_deployment(context, cluster, ng)
        machine_deployments.append(machine_deployment)

    return machine_deployments


class Cluster(ClusterBase):
    def __init__(
        self,
        context: context.RequestContext,
        api: magnum_cluster_api.KubeClient,
        pykube_api: pykube.HTTPClient,
        cluster: magnum_objects.Cluster,
        namespace: str = "magnum-system",
    ):
        self.context = context
        self.api = api
        self.pykube_api = pykube_api
        self.cluster = cluster
        self.namespace = namespace

    @property
    def api_version(self) -> str:
        return "cluster.x-k8s.io/v1beta1"

    @property
    def kind(self) -> str:
        return "Cluster"

    @property
    def name(self) -> str:
        return self.cluster.stack_id

    @property
    def labels(self) -> dict:
        labels = {}
        if self.cluster.cluster_template.network_driver == "calico":
            cni_version = self.cluster.labels.get("calico_tag", CALICO_TAG)
            labels = {
                "cni": f"calico-{cni_version}",
            }

        return {**super().labels, **labels}

    def get_or_none(self) -> objects.Cluster:
        return objects.Cluster.objects(
            self.pykube_api, namespace=self.namespace
        ).get_or_none(name=self.cluster.stack_id)

    def _get_admission_control_list(self) -> str:
        """Get admission control list with NodeRestriction as base.

        This matches the behavior of the Heat templates where NodeRestriction
        is always prepended to any user-provided admission plugins.
        """
        user_plugins = self.cluster.labels.get("admission_control_list", "")
        if user_plugins:
            return f"NodeRestriction,{user_plugins}"
        else:
            return "NodeRestriction"

    def get_object(self) -> dict:
        osc = clients.get_openstack_api(self.context)
        default_volume_type = osc.cinder().volume_types.default()
        pod_cidr = DEFAULT_POD_CIDR
        if self.cluster.cluster_template.network_driver == "calico":
            pod_cidr = self.cluster.labels.get(
                "calico_ipv4pool",
                DEFAULT_POD_CIDR,
            )
        if self.cluster.cluster_template.network_driver == "cilium":
            pod_cidr = self.cluster.labels.get(
                "cilium_ipv4pool",
                DEFAULT_POD_CIDR,
            )

        # Lookup the flavor from Nova
        control_plane_flavor = utils.lookup_flavor(osc, self.cluster.master_flavor_id)
        worker_flavor = utils.lookup_flavor(osc, self.cluster.flavor_id)
        image = utils.lookup_image(osc, self.cluster.default_ng_master.image_id)

        return {
            "metadata": {
                "labels": self.labels,
            },
            "spec": {
                "clusterNetwork": {
                    "serviceDomain": self.cluster.labels.get(
                        "dns_cluster_domain", "cluster.local"
                    ),
                    "pods": {
                        "cidrBlocks": [pod_cidr],
                    },
                    "services": {
                        "cidrBlocks": [
                            self.cluster.labels.get(
                                "service_cluster_ip_range", "10.254.0.0/16"
                            )
                        ],
                    },
                },
                "topology": {
                    "class": CLUSTER_CLASS_NAME,
                    "version": utils.get_kube_tag(self.cluster),
                    "controlPlane": {
                        "metadata": {
                            "labels": {
                                "node-role.kubernetes.io/master": "",
                            }
                        },
                        "replicas": self.cluster.master_count,
                        "machineHealthCheck": {
                            "enable": utils.get_auto_healing_enabled(self.cluster)
                        },
                    },
                    "workers": {
                        "machineDeployments": generate_machine_deployments_for_cluster(
                            self.context, self.cluster
                        ),
                    },
                    "variables": [
                        {
                            "name": "apiServerLoadBalancer",
                            "value": {
                                "enabled": self.cluster.master_lb_enabled,
                                "provider": self.cluster.labels.get(
                                    "octavia_provider", "amphora"
                                ),
                            },
                        },
                        {
                            "name": "apiServerTLSCipherSuites",
                            "value": self.cluster.labels.get(
                                "api_server_tls_cipher_suites",
                                "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305,TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305",  # noqa: E501
                            ),
                        },
                        {
                            "name": "openidConnect",
                            "value": {
                                "clientId": self.cluster.labels.get(
                                    "oidc_client_id", ""
                                ),
                                "groupsClaim": self.cluster.labels.get(
                                    "oidc_groups_claim", ""
                                ),
                                "groupsPrefix": self.cluster.labels.get(
                                    "oidc_groups_prefix", ""
                                ),
                                "issuerUrl": self.cluster.labels.get(
                                    "oidc_issuer_url", ""
                                ),
                                "usernameClaim": self.cluster.labels.get(
                                    "oidc_username_claim", "sub"
                                ),
                                "usernamePrefix": self.cluster.labels.get(
                                    "oidc_username_prefix", "-"
                                ),
                            },
                        },
                        {
                            "name": "auditLog",
                            "value": {
                                "enabled": utils.get_cluster_label_as_bool(
                                    self.cluster, "audit_log_enabled", False
                                ),
                                "maxAge": self.cluster.labels.get(
                                    "audit_log_max_age", "30"
                                ),
                                "maxBackup": self.cluster.labels.get(
                                    "audit_log_max_backup", "10"
                                ),
                                "maxSize": self.cluster.labels.get(
                                    "audit_log_max_size", "100"
                                ),
                            },
                        },
                        {
                            "name": "bootVolume",
                            "value": {
                                "size": utils.get_cluster_label_as_int(
                                    self.cluster,
                                    "boot_volume_size",
                                    CONF.cinder.default_boot_volume_size,
                                ),
                                "type": self.cluster.labels.get(
                                    "boot_volume_type",
                                    cinder.get_default_boot_volume_type(self.context),
                                ),
                            },
                        },
                        {
                            "name": "clusterIdentityRefName",
                            "value": utils.get_cluster_api_cloud_config_secret_name(
                                self.cluster
                            ),
                        },
                        {
                            "name": "systemdProxyConfig",
                            "value": base64.encode_as_text(
                                utils.generate_systemd_proxy_config(self.cluster)
                            ),
                        },
                        {
                            "name": "aptProxyConfig",
                            "value": base64.encode_as_text(
                                utils.generate_apt_proxy_config(self.cluster)
                            ),
                        },
                        {
                            "name": "containerdConfig",
                            "value": base64.encode_as_text(
                                utils.generate_containerd_config(self.cluster)
                            ),
                        },
                        {
                            "name": "controlPlaneFlavor",
                            "value": control_plane_flavor.name,
                        },
                        {
                            "name": "disableAPIServerFloatingIP",
                            "value": utils.get_cluster_floating_ip_disabled(
                                self.cluster
                            ),
                        },
                        {
                            "name": "dnsNameservers",
                            "value": self.cluster.cluster_template.dns_nameserver.split(
                                ","
                            ),
                        },
                        {
                            "name": "externalNetworkId",
                            "value": neutron.get_external_network_id(
                                self.context,
                                self.cluster.cluster_template.external_network_id,
                            ),
                        },
                        {
                            "name": "fixedNetworkId",
                            "value": utils.get_fixed_network_id(
                                self.context, self.cluster.fixed_network
                            )
                            or "",
                        },
                        {
                            "name": "fixedSubnetId",
                            "value": neutron.get_fixed_subnet_id(
                                self.context, self.cluster.fixed_subnet
                            )
                            or "",
                        },
                        {
                            "name": "flavor",
                            "value": worker_flavor.name,
                        },
                        {
                            "name": "imageRepository",
                            "value": utils.get_cluster_container_infra_prefix(
                                self.cluster,
                            ),
                        },
                        {
                            "name": "imageUUID",
                            "value": image.get("id"),
                        },
                        {
                            "name": "kubeletTLSCipherSuites",
                            "value": self.cluster.labels.get(
                                "kubelet_tls_cipher_suites",
                                "TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256,TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384,TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305",  # noqa: E501
                            ),
                        },
                        {
                            "name": "apiServerSANs",
                            "value": utils.generate_api_cert_san_list(self.cluster),
                        },
                        {
                            "name": "nodeCidr",
                            "value": self.cluster.labels.get(
                                "fixed_subnet_cidr",
                                "10.0.0.0/24",
                            ),
                        },
                        {
                            "name": "sshKeyName",
                            "value": self.cluster.keypair or "",
                        },
                        {
                            "name": "operatingSystem",
                            "value": utils.get_operating_system(self.cluster),
                        },
                        {
                            "name": "hardwareDiskBus",
                            "value": image.get("hw_disk_bus", ""),
                        },
                        {
                            "name": "enableDockerVolume",
                            "value": self.cluster.docker_volume_size is not None,
                        },
                        {
                            "name": "dockerVolumeSize",
                            "value": self.cluster.docker_volume_size or 0,
                        },
                        {
                            "name": "dockerVolumeType",
                            "value": self.cluster.labels.get(
                                "docker_volume_type",
                                default_volume_type.name,
                            ),
                        },
                        {
                            "name": "enableEtcdVolume",
                            "value": utils.get_cluster_label_as_int(
                                self.cluster,
                                "etcd_volume_size",
                                0,
                            )
                            > 0,
                        },
                        {
                            "name": "etcdVolumeSize",
                            "value": utils.get_cluster_label_as_int(
                                self.cluster,
                                "etcd_volume_size",
                                0,
                            ),
                        },
                        {
                            "name": "etcdVolumeType",
                            "value": self.cluster.labels.get(
                                "etcd_volume_type",
                                default_volume_type.name,
                            ),
                        },
                        {
                            "name": "availabilityZone",
                            "value": self.cluster.labels.get("availability_zone", ""),
                        },
                        {
                            "name": "enableKeystoneAuth",
                            "value": utils.get_cluster_label_as_bool(
                                self.cluster, "keystone_auth_enabled", True
                            ),
                        },
                        {
                            "name": "controlPlaneAvailabilityZones",
                            "value": self.cluster.labels.get(
                                "control_plane_availability_zones", ""
                            ).split(","),
                        },
                        # NOTE(oleks): Set cluster-level variable using server group id for controlplane.
                        #              Override this for node groups via  MachineDeployment-level variable.
                        {
                            "name": "serverGroupId",
                            "value": utils.ensure_controlplane_server_group(
                                ctx=self.context, cluster=self.cluster
                            ),
                        },
                        # NOTE(oleks): Set cluster-level variable using cluster label for controlplane.
                        #              Override this using node group label for node groups via  MachineDeployment-level variable. # noqa: E501
                        {
                            "name": "isServerGroupDiffFailureDomain",
                            "value": utils.is_controlplane_different_failure_domain(
                                cluster=self.cluster
                            ),
                        },
                        {
                            "name": "admissionControlList",
                            "value": self._get_admission_control_list(),
                        },
                    ],
                },
            },
        }

    def delete(self):
        capi_cluster = self.get_or_none()
        if capi_cluster:
            capi_cluster.delete()


def apply_cluster_from_magnum_cluster(
    context: context.RequestContext,
    api: magnum_cluster_api.KubeClient,
    pykube_api: pykube.HTTPClient,
    cluster: magnum_objects.Cluster,
    skip_auto_scaling_release: bool = False,
) -> None:
    """
    Create a ClusterAPI cluster given a Magnum Cluster object.
    """

    ClusterServerGroups(context, cluster).apply()
    LegacyClusterResourcesSecret(context, api, pykube_api, cluster).apply()
    CloudProviderClusterResourcesSecret(context, api, pykube_api, cluster).apply()

    if not skip_auto_scaling_release and utils.get_auto_scaling_enabled(cluster):
        ClusterAutoscalerHelmRelease(api, cluster).apply()


def get_kubeadm_control_plane(
    api: pykube.HTTPClient, cluster: magnum_objects.Cluster
) -> typing.Optional[objects.KubeadmControlPlane]:
    kcps = objects.KubeadmControlPlane.objects(api, namespace="magnum-system").filter(
        selector={
            "cluster.x-k8s.io/cluster-name": cluster.stack_id,
        },
    )
    if len(kcps) == 1:
        return list(kcps)[0]
    return None



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from eventlet import tpool  # type: ignore
from magnum.conductor import monitors  # type: ignore
from oslo_log import log as logging  # type: ignore

from magnum_cluster_api import clients, objects, utils
from magnum_cluster_api.magnum_cluster_api import Monitor as RustMonitor

LOG = logging.getLogger(__name__)


class Monitor(monitors.MonitorBase):
    def metrics_spec(self):
        pass

    def pull_data(self):
        pass

    def poll_nodegroup_replicas(self):
        """
        Poll the number of replicas of each nodegroup in the cluster when autoscaling enabled.
        """
        k8s_api = clients.get_pykube_api()
        if not utils.get_auto_scaling_enabled(self.cluster):
            return
        for node_group in self.cluster.nodegroups:
            md = objects.MachineDeployment.for_node_group_or_none(
                k8s_api, self.cluster, node_group
            )
            if md is None:
                continue
            node_group.node_count = md.obj["spec"]["replicas"]
            node_group.save()

    def poll_health_status(self):
        rust_monitor = tpool.Proxy(RustMonitor(self.cluster))
        self.data = rust_monitor.poll_health_status()

        self.poll_nodegroup_replicas()



# Copyright (c) 2023 VEXXHOST, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from magnum import objects
from magnum.common import context, exception
from oslo_log import log as logging

from magnum_cluster_api import clients, utils

LOG = logging.getLogger(__name__)


def is_enabled(
    cluster: objects.Cluster,
    label_flag: str,
    service_name: str,
) -> bool:
    return utils.get_cluster_label_as_bool(
        cluster, label_flag, True
    ) and is_service_enabled(service_name)


def is_service_enabled(service_type: str) -> bool:
    """Check if service is deployed in the cloud."""

    admin_context = context.make_admin_context()
    osc = clients.get_openstack_api(admin_context)
    keystone = osc.keystone()

    try:
        service = keystone.client.services.list(type=service_type)
    except Exception:
        LOG.exception("Failed to list services")
        raise exception.ServicesListFailed()

    if service and service[0].enabled:
        return True

    LOG.info("There is no %s service enabled in the cloud.", service_type)
    return False
